@@ -15,6 +15,8 @@
 # under the License.
 
 import datetime
+from random import Random
+import retrying
 
 from novaclient import exceptions as nova_exceptions
 from oslo_config import cfg
@@ -64,6 +66,9 @@ plugin_opts = [
                help='Interval (minutes) of reservation healing. '
                     'If 0 is specified, the interval is infinite and all the '
                     'reservations in the future is healed at one time.'),
+    cfg.BoolOpt('randomize_host_selection',
+                default=False,
+                help='Allocate hosts for reservations randomly.'),
 ]
 
 CONF = cfg.CONF
@@ -72,6 +77,7 @@ LOG = logging.getLogger(__name__)
 
 before_end_options = ['', 'snapshot', 'default']
 
+INSTANCE_DELETION_TIMEOUT = 10 * 60 * 1000  # 10 minutes
 QUERY_TYPE_ALLOCATION = 'allocation'
 
 
@@ -193,10 +199,11 @@ class PhysicalHostPlugin(base.BasePlugin, nova.NovaClientWrapper):
     def on_end(self, resource_id):
         """Remove the hosts from the pool."""
         host_reservation = db_api.host_reservation_get(resource_id)
+        reservation_id = host_reservation['reservation_id']
         db_api.host_reservation_update(host_reservation['id'],
                                        {'status': 'completed'})
         allocations = db_api.host_allocation_get_all_by_values(
-            reservation_id=host_reservation['reservation_id'])
+            reservation_id=reservation_id)
         for allocation in allocations:
             db_api.host_allocation_destroy(allocation['id'])
         pool = nova.ReservationPool()
@@ -210,11 +217,35 @@ class PhysicalHostPlugin(base.BasePlugin, nova.NovaClientWrapper):
                              'concurrently.', server)
                 except Exception as e:
                     LOG.exception('Failed to delete %s: %s.', server, str(e))
+
+        # We need to check the deletion is complete before removing the host
+        # from the aggregate. See change
+        # https://review.opendev.org/c/openstack/nova/+/821423 for details.
+        if not self._check_server_deletion(pool, host_reservation):
+            LOG.error('Timed out while deleting servers on reservation %s',
+                      reservation_id)
+            raise manager_ex.ServerDeletionTimeout()
+
         try:
             pool.delete(host_reservation['aggregate_id'])
         except manager_ex.AggregateNotFound:
             pass
 
+    @retrying.retry(stop_max_delay=INSTANCE_DELETION_TIMEOUT,
+                    wait_fixed=5000,  # 5 seconds interval
+                    retry_on_result=lambda x: x is False)
+    def _check_server_deletion(self, pool, host_reservation):
+        servers = []
+        for host in pool.get_computehosts(host_reservation['aggregate_id']):
+            servers.extend(
+                self.nova.servers.list(search_opts={"host": host,
+                                                    "all_tenants": 1},
+                                       detailed=False))
+        if servers:
+            LOG.info('Waiting to delete servers: %s ', servers)
+            return False
+        return True
+
     def heal_reservations(self, failed_resources, interval_begin,
                           interval_end):
         """Heal reservations which suffer from resource failures.
@@ -394,6 +425,34 @@ class PhysicalHostPlugin(base.BasePlugin, nova.NovaClientWrapper):
                 raise manager_ex.CantAddExtraCapability(
                     keys=cantaddextracapability,
                     host=host['id'])
+
+            # Check for placement details
+            hostname = host_details['hypervisor_hostname']
+            rp = self.placement_client.get_resource_provider(hostname)
+            if rp is None:
+                raise manager_ex.ResourceProviderNotFound(host=hostname)
+
+            inventories = self.placement_client.get_inventory(rp['uuid'])
+            for rc, inventory in inventories['inventories'].items():
+                resource_inventory = {
+                    'computehost_id': host['id'],
+                    'resource_class': rc,
+                    'total': inventory['total'],
+                    'reserved': inventory['reserved'],
+                    'min_unit': inventory['min_unit'],
+                    'max_unit': inventory['max_unit'],
+                    'step_size': inventory['step_size'],
+                    'allocation_ratio': inventory['allocation_ratio'],
+                }
+                db_api.host_resource_inventory_create(resource_inventory)
+
+            traits = self.placement_client.get_traits(rp['uuid'])
+            for trait in traits:
+                db_api.host_trait_create({
+                    'computehost_id': host['id'],
+                    'trait': trait,
+                })
+
             return self.get_computehost(host['id'])
 
     def is_updatable_extra_capability(self, capability, property_name):
@@ -593,9 +652,13 @@ class PhysicalHostPlugin(base.BasePlugin, nova.NovaClientWrapper):
             ]:
                 allocated_host_ids.append(host['id'])
         if len(not_allocated_host_ids) >= int(min_host):
+            if CONF[self.resource_type].randomize_host_selection:
+                Random.shuffle(not_allocated_host_ids)
             return not_allocated_host_ids[:int(max_host)]
         all_host_ids = allocated_host_ids + not_allocated_host_ids
         if len(all_host_ids) >= int(min_host):
+            if CONF[self.resource_type].randomize_host_selection:
+                Random.shuffle(all_host_ids)
             return all_host_ids[:int(max_host)]
         else:
             return []
